package net.sf.banglapad;

import net.sf.banglapad.UserKeyMap;

import java.io.*;
import java.awt.event.*;
import java.text.BreakIterator;
import javax.swing.text.*;
import javax.swing.JComboBox;
import java.util.*;

/**
 * BanglaKeyListener 
 * 
 * @author Naushad UzZaman (naushadzaman@gmail.com)
 * @author Zahurul Islam (zahurul@bracuniversity.net)
 * 
 * @version 0.1, September 21, 2005
 * 
 */

public class BanglaKeyListener extends KeyAdapter {
    private JTextComponent textComponent;
    
    //private static InputMethods inputMethod = InputMethods.Telex; // default keyboard layout
    
    private static boolean smartMarkOn;
    
    private static boolean banglaModeOn = true;
    
    private static boolean paliSanskritDegaOn;
    
    private int start, end;
    
    private String curWord, vietWord;
    
    private char curChar, vietChar;
    
    private char keyChar, accent;
    
    private static Properties macroMap;
    
    private static boolean repeatKeyConsumed;
    
    private static final char ESCAPE_CHAR = '\\';
    
    private static final String SHIFTING_CHARS = "cmnpt"; // chars potentially cause shifting of diacritical marks
    
    private static final String VOWELS = "aeiouy";
    
    private static final String NON_ACCENTS = "!@#$%&)_={}[]|:;/>,";
    
    private static final int MODIFIER_MASK = InputEvent.META_DOWN_MASK
    | InputEvent.CTRL_DOWN_MASK | InputEvent.ALT_DOWN_MASK;
    
    private static final BreakIterator boundary = BreakIterator
    .getWordInstance();
    
    private static char typed_key = 0; // tajpita klavo (signo) // from simredo
    
    private static char previously_typed = 0; // antauxe tajpita signo //from simredo
    
    /**
     * Creates a new <CODE>VietKeyListener</CODE>.
     *
     * @param textComponent the JTextComponent to be listened
     */
    public BanglaKeyListener(JTextComponent textComponent) {
        this.textComponent = textComponent;
        //textComponent.addKeyListener(new VietKeyListener(textComponent));
    }
    
    /**
     * Creates a new <CODE>VietKeyListener</CODE>.
     *
     * @param comboBox the JComboBox to be listened
     */
    public BanglaKeyListener(JComboBox comboBox) {
        JTextComponent textComponent = (JTextComponent) comboBox.getEditor()
        .getEditorComponent();
        textComponent.addKeyListener(new BanglaKeyListener(textComponent));
    }
    
    /**
     * Sets Vietnamese mode for key input.
     *
     * @param mode true to enable entry of Vietnamese characters (enabled by default)
     */
    public static void setVietModeEnabled(final boolean mode) {
        banglaModeOn = mode;
    }
    
    /**
     * Sets Pali-Sanskrit/Dega mode for key input.
     * <p>
     * This capability is dependent on VIQR input method being selected.
     *
     * @param mode true to enable entry of Pali-Sanskrit/Dega characters in addition to Vietnamese
     */
    public static void setPaliSanskritModeEnabled(final boolean mode) {
        paliSanskritDegaOn = mode;
    }
    
    
    
    /**
     * Sets the SmartMark capability on or off.
     *
     * @param smartMark true to enable automatic placement of diacritical marks on appropriate vowels in a word;<br>
     * otherwise, they must be typed immediately after the character they qualify.
     */
    public static void setSmartMark(final boolean smartMark) {
        smartMarkOn = smartMark;
    }
    
    
    /**
     * Sets map for shorthand.
     *
     * @param shortHandMap the list of shorthand sequences
     */
    public static void setMacroMap(final Properties shortHandMap) {
        macroMap = shortHandMap;
    }
    
    /**
     * FIXME: keyAction is called in the KeyTyped listener. It is called when
     * a key is typed. Kept the action for Bangla key typed in a specific
     * method, so that it can be changed and modified easily. 
     * @param e is the KeyEvent found from KeyTyped
     */
    public void keyAction(KeyEvent e) {
        BPadLogger.logger.info("KeyAction is called from KeyTyped");
        int pos; // Position of caret. Pozicio de tekstomontrilo.
        String previous_chars;
        char accented_character;
        String replacement;
        int the_key_code = e.getKeyCode();
        int the_key_id = e.getID();
        
        char[] typed_key_array = { 0 };
        typed_key_array[0] = e.getKeyChar();
        typed_key = e.getKeyChar();
        BPadLogger.logger.info("Checking if UserKeyMap is enabled");
        if (UserKeyMap.isEnabled()) {
            BPadLogger.logger.info("UserKeyMap is enabled");
            // Key maps are such a bother...
            
            if (the_key_id == KeyEvent.KEY_PRESSED
                    || the_key_id == KeyEvent.KEY_RELEASED)
                return;
            
            // Try to translate deadkey. Provu traduki senpasxan klavon.
            replacement = UserKeyMap.translate(previously_typed, typed_key);
            if (replacement != null) {
                //if (the_key_id == KeyEvent.KEY_TYPED) {
                pos = textComponent.getCaretPosition();
                textComponent.select(pos - 1, pos);
                textComponent.replaceSelection(replacement);
                previously_typed = replacement.charAt(0);
                //}
                e.consume();
                return;
            }
            BPadLogger.logger.info("From the user key map trying to extract the replacement key, i.e. KA for k");
            replacement = UserKeyMap.translate(typed_key, e.isControlDown(), e.isAltDown());
            if (replacement != null) {
                BPadLogger.logger.info("Got the replacement key for:"+typed_key+" and it is:"+replacement);
                textComponent.replaceSelection(replacement);
                previously_typed = replacement.charAt(0);
                e.consume();
                return;
            }
            super.keyTyped(e);
            if (typed_key >= '!')
                previously_typed = typed_key;
        } else { // user key map not active  /  Klavmapo neaktiva.
            super.keyTyped(e);
        }
    }
    
    /**
     * FIXME: Fully modified for BanglaKeyListener. Now it doesn't use Vietnamese specific keyListener.   
     */
    /**
     * Invoked when a key has been typed.
     *
     * @param e invoked when a key has been typed. This event occurs when a key press is followed by a key release.
     */
    public void keyTyped(KeyEvent e) {
        BPadLogger.logger.debug("Key has been typed");
        keyAction(e);
    }
    
    private String getCurrentWord(int pos, String source) {
        //        BreakIterator boundary = BreakIterator.getWordInstance();
        boundary.setText(source);
        end = boundary.following(pos - 1);
        start = boundary.previous();
        end = pos; // fine-tune word end
        return source.substring(start, end);
    }
}
